
import numpy as np
import matplotlib.pyplot as plt


grid = 32;

gravit = 0.0000001;

Laplacian = np.zeros((grid,grid), dtype = complex );
invLaplacian = np.zeros((grid,grid), dtype = complex );
formWeights = np.zeros((grid,grid), dtype = complex );
invFormWeights = np.zeros((grid,grid), dtype = complex );

x = np.linspace(0,2*np.pi, grid, endpoint = False ); 
y = np.linspace(0,2*np.pi, grid, endpoint = False ); 
x1,x2 = np.meshgrid(x,y);
freqs = np.fft.fftfreq( grid, 1./grid);
freqs = np.fft.fftshift(freqs).astype(complex);

zetaCoeffs = np.zeros((grid,grid), dtype = complex );
divCoeffs = np.zeros((grid,grid), dtype = complex );
hCoeffs = np.zeros((grid,grid), dtype = complex );
zetaCoeffs_lt = np.zeros((grid,grid), dtype = complex );
divCoeffs_lt = np.zeros((grid,grid), dtype = complex );
hCoeffs_lt  = np.zeros((grid,grid), dtype = complex );                    
 
dotzetaCoeffs = np.zeros((grid,grid), dtype = complex );
dotdivCoeffs = np.zeros((grid,grid), dtype = complex );
dothCoeffs = np.zeros((grid,grid), dtype = complex );                      

#arrays here initialized with natural indexing
invFormWeights[np.int(grid/2), np.int(grid/2)] = 1.
formWeights[np.int(grid/2), np.int(grid/2)] = 1.
               
for k1 in range(0,grid):
    for k2 in range(0, grid):
        w = (freqs[k1]*freqs[k1] + freqs[k2]*freqs[k2] )        
        
        Laplacian[k1,k2] = - w;
        formWeights[k1,k2] = np.sqrt(w); 
        if k1 != grid/2 and k2!=grid/2:
            invLaplacian[k1,k2] = - 1./w;
            invFormWeights[k1,k2] = 1./np.sqrt(w) ;                  
                     


def loadInitialConditions():
    """
    Load main variables with initial values

    """
    
    global zetaCoeffs, divCoeffs, hCoeffs
    
    zetaInit = 0.0001 * np.sin(x2);
    hInit = 0.0001 + 0.0001 * np.sin(x2);
                 
    zetaCoeffs = np.fft.fftshift( np.fft.fft2(zetaInit))/(grid*grid);                                   
    hCoeffs = np.fft.fftshift( np.fft.fft2(hInit))/(grid*grid);                           
                             
def reconstructForm(Type: np.int, fCoeff: np.ndarray):
    """
    evaluates 0 and 2 forms in the spacial domain
    
    Arguments:
    Type: degree of the differential form to be evaluated (type: np.int)   
    fCoeffs: fourier coefficients in natural indexing (type: np.ndarray, size: (grid,grid))
    
    Return:
    f: evaluation of the differential form in the uniform grid (type: np.ndarray, size(grid,grid))
    
    Raised Exceptions:
    ValueError: if passed type is neither 0 nor 2


    """
    
    if Type != 0 and Type != 2:
        raise ValueError('Type must be 0 or 2.')


    fCoeffs =np.copy(fCoeff);
    
    if Type == 0:
        #using the weights to evaluate forms
        fCoeffs *= invFormWeights;
        #scaling coefficients to fit ifft2
        fCoeffs *= grid*grid;
        #rearranging coefficients to fit ifft2
        fCoeffs = np.fft.ifftshift(fCoeffs);              
        f =  np.fft.ifft2(fCoeffs);
                       
    if Type ==2:
        #using the weights to evaluate forms
        fCoeffs *= formWeights;
        #scaling coefficients to fit ifft2
        fCoeffs *= grid*grid;
        #rearranging coefficients to fit ifft2
        fCoeffs = np.fft.ifftshift(fCoeffs);              
        f =  np.fft.ifft2(fCoeffs);
                         
    return f                     
        
def reconstruct1Form(Type: np.int, fCoeff: np.ndarray):
    """
    evaluates 1 forms in the spacial domain
    
    Arguments:
    Type: exact (==0) or coexact (==1) differential 1-form to be evaluated (type: np.int)   
    fCoeffs: fourier coefficients in natural indexing (type: np.ndarray, size: (grid,grid))
    
    Return:
    f: evaluation of the differential form in the uniform grid (type: np.ndarray, size(grid,grid))
    
    Raised Exceptions:
    ValueError: if passed type is neither 0 nor 1

    """
    
    if Type != 0 and Type != 1:
        raise ValueError('Type must be 0 or 1.')

    fCoeffs =np.copy(fCoeff);    
    
    #scaling coefficients to fit ifft2
    fCoeffs *= grid*grid;
    #spliting coeffs for each entry                                  
    fCoeffs1 = np.copy(fCoeffs);
    fCoeffs2 = np.copy(fCoeffs);

    for k in range(0,grid):
        fCoeffs1[k, :] *= (1.j)*freqs;
        fCoeffs2[:, k] *= (1.j)*freqs;
        
    #rearranging coefficients to fit ifft2
    fCoeffs1 = np.fft.ifftshift(fCoeffs1);              
    fCoeffs2 = np.fft.ifftshift(fCoeffs2);             

    fx1 =  np.fft.ifft2(fCoeffs1);
    fx2 =  np.fft.ifft2(fCoeffs2);                   
    if Type == 0:        
        f = np.array([fx1, fx2]);                           
    if Type ==1:
        f = np.array([-fx2,fx1]);     
                 
    return f                     

    
def projectForm(Type: np.int, sig: np.ndarray):
    """
    projects 0 and 2 forms on the frequency domain
    
    Arguments:
    Type: degree of the differential form to be projected (type: np.int)   
    f: evaluation of the differential form in the uniform grid (type: np.ndarray, size(grid,grid))
    
    Return:
    fCoeffs: fourier coefficients in natural indexing (type: np.ndarray, size: (grid,grid))
    
    Raised Exceptions:
    ValueError: if passed type is neither 0 nor 2

    """
    
    if Type != 0 and Type != 2:
        raise ValueError('Type must be 0 or 2.')

    f = np.copy(sig); 
    
    if Type == 0:
        #scaling the function to fit fft2
        f /= grid*grid;
        #projecting 
        fCoeffs = np.fft.fft2(f);              
        #rearranging coefficients to natural indexing
        fCoeffs = np.fft.fftshift(fCoeffs);              
        #using the weights to evaluate forms
        fCoeffs *= formWeights;
               
    if Type ==2:
        #scaling the function to fit fft2
        f /= grid*grid;
        #projecting
        fCoeffs = np.fft.fft2(f);              
        #rearranging coefficients to natural indexing
        fCoeffs = np.fft.fftshift(fCoeffs);              
        #using the weights to evaluate forms
        fCoeffs *= invFormWeights;
                         
    return fCoeffs                     
   
    
  
def infinitesimalAdvection(idx: np.int):
    """
    Calculates time derivatives of the main variables given their current values

    Arguments:
    zetaCoeffs : Fourier coefficients of the vorticity 2-form (type: np.ndarray, size: (L+1))
    divCoeffs: Fourier coefficients of the divergence 2-form (type: np.darray, size: (L+1))
    hCoeffs: Fourier coefficients of the height 0-form (type: np.darray, size: (L+1))    
    
    Return:
    dotzetaCoeffs : Fourier coefficients of the time derivative of the vorticity 2-form (type: np.ndarray, size: (L+1))
    dotdivCoeffs: Fourier coefficients of the time derivative of the divergence 2-form (type: np.darray, size: (L+1))
    dothCoeffs: Fourier coefficients of the time derivative of the height 0-form (type: np.darray, size: (L+1))
    
    all the coefficient arrays are indexed with natural indexing
    """
    
    global zetaCoeffs, divCoeffs, hCoeffs, dotzetaCoeffs, dotdivCoeffs, dothCoeffs
    
    #calculating velocity vector field in the spacial domain
    
    u_dd = reconstruct1Form(0, divCoeffs);
    u_delta = reconstruct1Form(1, zetaCoeffs);                       
    u = u_dd + u_delta;

          
    # compute coefficients for vorticity derivative     
                

    # compute coefficients for divergence derivative            
                                   

    # compute coefficients for height derivative


    
    # write out image files (for video generation)
        
    #fig1, ax1 = plt.subplots( nrows=1, ncols=1 ) 
    #plt.quiver(x1,x2,np.real(u[0]),np.real(u[1]) )
    #fig1.savefig( './images/re_velocity_' + "{:04}".format(idx))
    #plt.close(fig1)
    
    #fig2, ax2 = plt.subplots( nrows=1, ncols=1 ) 
    #plt.quiver(x1,x2,np.imag(u[0]),np.imag(u[1]))
    #fig2.savefig( './images/im_velocity_' + "{:04}".format(idx))
    #plt.close(fig2)                        
    
    # video can be generated on the command line with
    # ffmpeg -r 15 -f image2 -i re_velocity_0%*.png -vcodec libx264 -pix_fmt yuv420p re_velocity.mp4
    # ffmpeg -r 15 -f image2 -i im_velocity_0%*.png -vcodec libx264 -pix_fmt yuv420p im_velocity.mp4

 
def advectEuler(dt, idx):
    global zetaCoeffs, divCoeffs, hCoeffs, zetaCoeffs_lt, divCoeffs_lt, hCoeffs_lt
    global dotzetaCoeffs, dotdivCoeffs, dothCoeffs 
              
    infinitesimalAdvection(idx);
                          
    #storing values for the previous step                                           
    zetaCoeffs_lt = zetaCoeffs;
    divCoeffs_lt = divCoeffs;
    hCoeffs_lt = hCoeffs;                                          
    
    zetaCoeffs += dt * dotzetaCoeffs;                      
    divCoeffs += dt * dotdivCoeffs;
    hCoeffs += dt * dothCoeffs;
     

def advectLeapFrog(dt, idx):
    global zetaCoeffs, divCoeffs, hCoeffs, zetaCoeffs_lt, divCoeffs_lt, hCoeffs_lt
    global dotzetaCoeffs, dotdivCoeffs, dothCoeffs 

    
    #Robert time smoothing constant
    r = 0.02;
    zetaCoeffs_llt = np.copy(zetaCoeffs_lt);
    divCoeffs_llt = np.copy(divCoeffs_lt);                            
    hCoeffs_llt = np.copy(hCoeffs_lt);    
    
    infinitesimalAdvection(idx);                     
    
    temp = np.copy(zetaCoeffs);
    zetaCoeffs = np.copy(zetaCoeffs_lt);
    zetaCoeffs_lt = np.copy(temp);                         
    
    zetaCoeffs += 2.0 * dt * dotzetaCoeffs;
    zetaCoeffs_lt = (1 - 2*r) * zetaCoeffs_lt + r*(zetaCoeffs+zetaCoeffs_llt); 
    
    temp = np.copy(divCoeffs);
    divCoeffs = np.copy(divCoeffs_lt);
    divCoeffs_lt = np.copy(temp);                         
    
    divCoeffs += 2.0 * dt * dotdivCoeffs;
    divCoeffs_lt = (1 - 2*r) * divCoeffs_lt + r*(divCoeffs+divCoeffs_llt); 

    temp = np.copy(hCoeffs);
    hCoeffs = np.copy(hCoeffs_lt);
    hCoeffs_lt = np.copy(temp);                         
    
    hCoeffs += 2.0 * dt * dothCoeffs;
    hCoeffs_lt = (1 - 2*r) * hCoeffs_lt + r*(hCoeffs + hCoeffs_llt); 
                         
def run(dt, total_time):
    global zetaCoeffs, divCoeffs, hCoeffs, zetaCoeffs_lt, divCoeffs_lt, hCoeffs_lt
    global dotzetaCoeffs, dotdivCoeffs, dothCoeffs 

    loadInitialConditions();
    t = 0.0;
    idx = 0;
    print( t, " :: ", np.linalg.norm(zetaCoeffs), " :: ", np.linalg.norm(divCoeffs), " :: ", np.linalg.norm(hCoeffs) );
    
    advectEuler(dt, idx);     
    t += dt;
    idx += 1;     
    while t <= total_time :
        print( t, " :: ", np.linalg.norm(zetaCoeffs), " :: ", np.linalg.norm(divCoeffs), " :: ", np.linalg.norm(hCoeffs) );
        advectLeapFrog(dt, idx);
        t += dt;                                  
        idx += 1;  
    
    
run(18., 3600.);    
